unit Map;

interface

uses
  SysUtils, IdGlobal, Token, MyRobot, Packages;

type
  MapPackage = Record
    Id: Integer;                // Id of the Package
    Dest: Point;                // Destination Point
    Weight: Integer;            // Packages Weight
  end;

type
  MapObject = Record
    flag: Integer;
    robot: Integer;
    packages: Boolean;
    myPacks: TPackages;
  end;

type
  TMap = Class (TObject)
  public
    constructor Create(Data: String);
    destructor Destroy(); override;
    procedure ParseLine(Data: String);
    procedure ParseGameLine(var Data: String);
    procedure setPackageLine(Data: String);
    procedure ParsePackageLine();
    procedure CreateMyPlayer(Data: String);
    procedure DisplayMap();

    procedure robotAI();

    function haveAllData(): Boolean;
    function getPath(pStart, pEnd: Point): String;
    function amIAlone(dis: Integer): Boolean;
    function canDestroy(var dir: char): Boolean;
  private
    width, height: Integer;          // Width and Height of the map
    lineOn: Integer;
    amSetup: Boolean;

    // 2D array for our map Objects
    mapData: array of array of MapObject;

    // My Robot
    myRobot: TMyRobot;

    // Package Line
    packageLine: String;

    procedure ParsePlayerMove(Data: String);

    function findClosestPackage(pStart: Point; var path: String): Point;
  published
    property Setup: Boolean read amSetup;
    property me: TMyRobot read myRobot;
end;


implementation

const
  FLAG_UNKNOWN = $FF;
  FLAG_PLAIN = $00;
  FLAG_HOME = $01;
  FLAG_WATER = $10;
  FLAG_WALL = $11;

// Sets up our map space, Data is the first line and gives us
// the width and height of the map;
constructor TMap.Create(Data: String);
var
  i, j: Integer;
  myToken: TToken;
begin
  inherited Create();

  myToken := TToken.Create(Data, ' ', '+');
  if (myToken.countTokens = 2) then begin
    LineOn := 0;        // We haven't parsed any lines (map data) yet
    width := StrToInt(myToken.nextToken());
    height := StrToInt(myToken.nextToken());
    setLength(mapData, height, width);
    WriteLn('Parsing Map Data....');
    amSetup := true;
    // Decrease width and height by 1 so we don't have to do it in loops
    dec(height);
    dec(width);
    for i := 0 to height do begin
      for j := 0 to width do begin
        mapData[i][j].flag := FLAG_UNKNOWN;
        mapData[i][j].robot := -1;
        mapData[i][j].packages := false;
      end;
    end;
  end
  else begin
    WriteLn('Malformed Map Constructor, first line must be map width height');
    amSetup := false;
  end;
end;

// Destroys us and all the map data we have, must play nice with memory.
Destructor TMap.Destroy();
var i, j: Integer;
begin
  myRobot.Free();
  for i := 0 to height do begin
    for j := 0 to width do begin
      mapData[i][j].myPacks.Free();
    end;
  end;

  inherited Destroy();
end;

Procedure TMap.ParseLine(Data: String);
var
  i: Integer;
  c: Char;
begin
  // Make sure we're setup before blindly parsing the data
  if (amSetup) then begin
    if (length(Data) = width + 1) then begin
      for i := 0 to width do begin
        c := Data[i+1];
        case (c) of
          '.': MapData[lineOn][i].flag := FLAG_PLAIN;
          '@': begin
                 MapData[lineOn][i].flag := FLAG_HOME;
                 MapData[lineOn][i].packages := true;
               end;
          '~': MapData[lineOn][i].flag := FLAG_WATER;
          '#': MapData[lineOn][i].flag := FLAG_WALL;
        end;
      end;
      inc(lineOn);
    end
    else begin
      WriteLn('Malformed map data on line ' + IntToStr(lineOn) + '!!!!');
    end;
  end;
end;

// This just splits our data into individual player moves by splitting
// by the substring #.
procedure TMap.ParseGameLine(var Data: String);
var myToken: TToken;
begin
  // Remove starting #
  Data := Copy(Data, 2, length(Data));

  // Create the tokenizer and parse every player's move
  myToken := TToken.Create(Data, '#', '+');
  while (myToken.hasMoreTokens) do begin
    self.ParsePlayerMove(myToken.nextToken());
  end;
  myToken.free();
end;

procedure TMap.setPackageLine(Data: String);
begin
  packageLine := data;
end;

procedure TMap.ParsePackageLine();
var
  dest: Point;
  path: String;
  tempPacks: TPackages;
begin
  if (length(packageLine) > 0) then begin
    tempPacks := mapData[myRobot.location.y][myRobot.location.x].myPacks;
    if (tempPacks = nil) then
      tempPacks := TPackages.Create(packageLine)
    else
      tempPacks.buildPackageList(packageLine);
    mapData[myRobot.location.y][myRobot.location.x].packages := true;
    mapData[myRobot.location.y][myRobot.location.x].myPacks := tempPacks;

    // Now that we have the package list we need to find the distance
    // to the diliery points
    dest := tempPacks.getNextDest();
    while (dest.x >= 0) do begin
      path := getPath(myRobot.location, dest);
      if (path[1] = '!') then
        tempPacks.setDistance(dest, -1)
      else
        tempPacks.setDistance(dest, length(path));
      dest := tempPacks.getNextDest();
    end;
  end
  else begin
    mapData[myRobot.location.y][myRobot.location.y].packages := false;
  end;;
end;

procedure TMap.CreateMyPlayer(Data: String);
begin
  myRobot := TMyRobot.Create(Data);
end;

procedure TMap.DisplayMap();
var
  i, j: Integer;
  line: String;
begin
  if (amSetup) and (lineOn = height+1) then begin
    for i := 0 to height do begin
      line := '';
      for j := 0 to width do begin
        case (MapData[i][j].flag) of
          FLAG_PLAIN:   line := line + '.';
          FLAG_HOME:    line := line + '@';
          FLAG_WATER:   line := line + '~';
          FLAG_WALL:    line := line + '#';
          FLAG_UNKNOWN: line := line + '?';
        end;
      end;
      WriteLn(line);
    end;
  end;
end;

// This function does all the heavy lifting when it comes to
// performing AI functionality.
procedure TMap.robotAI();
var
  dest: Point;          // Destination Point
  path: String;         // Walking Path
  pList: String;        // Package List
  alone: Boolean;       // Am I alone?
  tempPacks: TPackages;
begin
  alone := amIAlone(2);
  if (myRobot.state = ROBOT_DELIVER) then begin
    if (alone) then
      myRobot.bid := 1;

    if (not myRobot.haveDest) then
      myRobot.setDest(myRobot.dest, myRobot.path)
    else if (myRobot.atDest) then
      myRobot.dropDelivery()
    else if (myRobot.onPath) then
      myRobot.onPathMove()
    else begin
      path := getPath(myRobot.location, myRobot.dest);
      myRobot.setDest(myRobot.dest, path);
    end;
  end
  else if (myRobot.state = ROBOT_FINDPACKAGE) then begin
    // If I'm alone no need to bid high;
    if (alone) then
      myRobot.bid := 1;

    if (not myRobot.haveDest) then begin
      // Get the point of the closest known package
      path := '';
      dest := findClosestPackage(myRobot.location, path);
      if (dest.x >= 0) and (dest.y >= 0) then begin
        myRobot.setDest(dest, path);
      end;
    end
    else begin
      // Check to see if we're at out dest, if so it's time to pick
      // up packages.
      if (myRobot.atDest) then begin
        writeLn('Time to pick up Packages...');
        tempPacks := mapData[myRobot.location.y][myRobot.location.x].myPacks;
        if (tempPacks <> nil) then begin
          pList := tempPacks.getBestPackageList(myRobot.capacity, dest);
          if (length(pList) > 0) then begin
            path := getPath(myRobot.location, dest);
            myRobot.setDelivery(dest, path, pList);
          end
          else begin
            mapData[myRobot.location.y][myRobot.location.x].packages := false;
            myRobot.haveDest := false;
            RobotAI();
          end;
        end
        // Uh oh, no packages where they should be, time to find another place
        else begin
          // We already waited and still nothing, find them elsewhere
          if (myRobot.waitOnPackage) then begin
          end

          // Wait one turn for them to come.
          else begin
            myRobot.move := 'Drop';
            myRobot.moveData := '-1';
          end;
        end;
      end
      // If the last move to our dest was on path, just keep on trucking
      else if (myRobot.onPath) then begin
        myRobot.onPathMove();
      end
      // If not, then we should once again find our closest package.
      else begin
        path := '';
        dest := findClosestPackage(myRobot.location, path);
        if (dest.x >= 0) and (dest.y >= 0) then begin
          myRobot.setDest(dest, path);
        end;
      end;
    end;
  end;
end;

// Returns true if we've read all our lines, false otherwise
function TMap.haveAllData(): Boolean;
begin
  Result := (lineOn = (height+1));
end;

// Trys to find the shortest path between the two points
//
// Works from the end point and building a movement cost
// array.  The movement cost at the start point is always
// the lowest value posible and a path will always be found.
//
// The only problem with this is that it can take quite some
// time to run on larger maps and it can eat a lot of stack space.
// To prevent a runaway loop, we'll limit our steps to the
// map width + map height.
function TMap.getPath(pStart, pEnd: Point): String;
var
  maxSteps, steps: Integer;     // Maximum number of steps to calculate
  markMap: array of array of Integer;   // Our Mark Map
  i, j: Integer;        // For counting through all map data
  found: Boolean;       // Makes sure we've found a next step
  sx, sy: Integer;      // Start x, start Y looping;
  ex, ey: Integer;      // end x, end y looping
  atX, atY: Integer;    // At X and At Y, for walking the path
  lowvalue: Integer;    // Lowest Distance
  noGo: Int64;        // Highest posible path length + 1
begin
  Result := '';
  maxSteps := (width+1) + (height+1);
  noGo := (width+1) * (height+1) + 1;
  setLength(markMap, height+1, width+1);
  { Fill up the mark Map }
  for i := 0 to height do begin
    for j := 0 to width do begin
      { Water, walls or robots cannot be passed }
      if (mapData[i][j].flag > $0F) or
         (mapData[i][j].robot <> -1) then
        markMap[i][j] := noGo
      else
        markMap[i][j] := -1;
    end;
  end;

  markMap[pEnd.y][pEnd.x] := 0;
  sx := pEnd.x;         // Set our start x to end point
  sy := pEnd.y;         // Set our start y to end point
  ex := pEnd.x;         // Set our end x to end point
  ey := pEnd.y;         // Set out end y to end point
  steps := 0;                   // Current number of steps
  found := true;

  // Find out path, keep adding
  while (found) and (steps < maxSteps) do begin
    found := false;
    for i := sy to ey do begin
      for j := sx to ex do begin
        // Only scan if our current location is passable and has a set value
        if (markMap[i][j] <> noGo) and (markMap[i][j] <> -1) then begin
          // Scan if Moving North is allowed, if so set mark map

          if (markMap[Max(i-1,0)][j] = -1) then begin
            found := true;
            markMap[Max(i-1,0)][j] := markMap[i][j] + 1;
          end;
          // Scan if Moving South is allowed, if so set mark map
          if (markMap[Min(i+1,height)][j] = -1) then begin
            found := true;
            markMap[Min(i+1,height)][j] := markMap[i][j] + 1;
          end;
          // Scan if Moving West is allowed, if so set mark map
          if (markMap[i][Max(j-1,0)] = -1) then begin
            found := true;
            markMap[i][Max(j-1,0)] := markMap[i][j] + 1;
          end;
          // Scan if Moving East is allowed, if so set mark map
          if (markMap[i][Min(j+1,width)] = -1) then begin
            found := true;
            markMap[i][Min(j+1,width)] := markMap[i][j] + 1;
          end;
        end;
      end;
    end;
    sx := Max(sx-1, 0);         // Set Our start x to start x - 1 or map edge
    sy := Max(sy-1, 0);         // Set out start y to start y - 1 or map edge
    ex := Min(ex+1, width);     // Set out end x to end y + 1 or map edge
    ey := Min(ey+1, height);    // Set out end y to end y + 1 or map edge
    inc(steps);
  end;

  // Now we have check to see if a path was found and build it
  atX := pStart.x;
  atY := pStart.y;
  lowValue := markMap[atY][atX];

  if (markMap[atY][atX] <> -1) and (markMap[atY][atX] <> noGo) then begin
    while (lowValue > 0) do begin
      // Check to see if moving North gets us closer to our end point
      if (atY-1 >= 0) and
         (markMap[atY-1][atX] < lowValue) then begin
        Result := Result + STMOVE_NORTH;
        atY := atY - 1;   // Move North
      end
      // Check to see if moving South gets us closer to our end point
      else if (atY+1 <= height) and
              (markMap[atY+1][atX] < lowValue) then begin
        Result := Result + STMOVE_SOUTH;
        atY := atY + 1;   // Move South
      end
      // Check to see if moving West gets us closer to out end point
      else if (atX-1 >= 0) and
              (markMap[atY][atX-1] < lowValue) then begin
        Result := Result + STMOVE_WEST;
        atX := atX - 1;   // Move West;
      end
      // Check to see if moving East gets us closer to out end point
      else if (atX+1 <= width) and
              (markMap[atY][atX+1] < lowValue) then begin
        Result := Result + STMOVE_EAST;
        atX := atX + 1;
      end;

      // Update our low value;
      lowValue := markMap[atY][atX];
    end;
  end
  else begin
    Result := '!!! No Path !!!';
  end;
end;

// This function determins if player's robot is alone (no one)
// within range of pushing no mater how they move.  If we are
// alone then our bid really doesn't matter 1 will be just fine.
function TMap.amIAlone(dis: Integer): Boolean;
var
  sx,sy: Integer;       // Starting x, y
  ex,ey: Integer;       // Ending x,y
  i, j: Integer;
  checkDis: Integer;    // Check Out Distance
  loc: Point;           // Player's Location
begin
  Result := true;
  if (amSetup and myRobot.ready) then begin
    loc := myRobot.location;
    sx := Max(loc.x-dis,0);
    sy := Max(loc.y-dis,0);
    ex := Min(loc.x+dis,width);
    ey := Min(loc.y+dis,height);
    for i := sy to ey do begin
      for j := sx to ex do begin
        // Distance check, no need to check if the distance is over 2
        // This oddly enough takes longer to compute but is more accurate
        checkDis := Abs(sx-loc.x) + Abs(sy-loc.y);
        // Check to see if there is another robot in range, if so, find it!
        if (mapData[i][j].robot > 0) and (checkDis <= dis) and
           (mapData[i][j].robot <> myRobot.identifier) then
          Result := false;
      end;
    end;
  end;
end;

// This function, much like amAlone, checks to see if we have
// the chance to destroy a player (push them into water).  If we
// can, the dir variable it set to the direction we need to move.
function TMap.canDestroy(var dir: Char): Boolean;
var
  loc: Point;           // Player's Location
begin
  Result := false;
  if (amSetup and myRobot.ready) then begin
    // Check a push to the north
    if (loc.y-2 >=0) and
       (mapData[loc.y-1][loc.x].robot > 0) and
       (mapData[loc.y-2][loc.x].flag = FLAG_WATER) then begin
      Result := true;
      dir := STMOVE_NORTH;
    end
    // Check a push to the south
    else if (loc.y+2 <= height) and
            (mapData[loc.y+1][loc.x].robot > 0) and
            (mapData[loc.y+2][loc.x].flag = FLAG_WATER) then begin
      Result := true;
      dir := STMOVE_SOUTH;
    end
    // Check a push to the west
    else if (loc.x-2 >= 0) and
            (mapData[loc.y][loc.x-1].robot > 0) and
            (mapData[loc.y][loc.x-2].flag = FLAG_WATER) then begin
      Result := true;
      dir := STMOVE_WEST;
    end
    // Check a push to the east
    else if (loc.x+2 <= width) and
            (mapData[loc.y][loc.x+1].robot > 0) and
            (mapData[loc.y][loc.x-2].flag = FLAG_WATER) then begin
      Result := true;
      dir := STMOVE_EAST;
    end;
  end;
end;


procedure TMap.ParsePlayerMove(Data: String);
var
  myToken: TToken;
  myBotId: Integer;
  move: Char;
begin
  myToken := TToken.Create(Data, ' ', '+');
  myBotId := StrToInt(myToken.nextToken());
  while (myToken.hasMoreTokens) do begin
    move := myToken.nextToken()[1];
    if (myBotId = myRobot.identifier) then begin
      try
        case (move) of
          'X':  myRobot.setLocationX(StrToInt(myToken.nextToken()));
          'Y':  myRobot.setLocationY(StrToInt(myToken.nextToken()));
          'S':  myRobot.doMove('S');
          'N':  myRobot.doMove('N');
          'E':  myRobot.doMove('E');
          'W':  myRobot.doMove('W');
        end;
      finally
      end;
    end;
  end;
  myToken.Free();
end;

function TMap.findClosestPackage(pStart: point; var path: String): Point;
var
  i,j: Integer;         // Counters
  pEnd: Point;          // End Point (Where the packages are)
  lowDis: Integer;      // Lowest Distance found so far
  tempPath: String;
begin
  Result.x := -1; Result.y := -1;
  lowDis := -1;
  for i := 0 to height do begin
    for j := 0 to width do begin
      if (mapData[i][j].packages) then begin
        pEnd.x := j;
        pEnd.y := i;
        tempPath := getPath(pStart, pEnd);
        if (tempPath[1] <> '!') then begin
          if (lowDis = -1) or (length(tempPath) < lowDis) then begin
            lowDis := length(tempPath);
            path := tempPath;
            Result := pEnd;
          end;
        end;
      end;
    end;
  end;
end;

end.
